"""
The :mod:`evopreprocess.feature_selection` provides a way to select features with evolutionary and nature inspired
algorithms.
"""

from .EvoFeatureSelection import EvoFeatureSelection

__all__ = ['EvoFeatureSelection']
