"""
The :mod:`evopreprocess.data_weighting` provides a way to weight instances in datasets with evolutionary and
nature inspired algorithms.
"""

from .EvoWeighting import EvoWeighting

__all__ = ['EvoWeighting']
